# -*- coding: utf-8 -*-

"""\
Tools I find myself constantly rebuilding and reusing.

I enjoy writing things myself from a relatively low level, so while some of
these classes are simply inferior to using a 'real' lightweight framework, I
find that I learn much much quicker doing things on my own, and that I also
love the feeling of understanding exactly how everything works in my projects.

Source: https://github.com/cmyui/cmyui_pkg

:copyright: (c) 2020 cmyui
:license: MIT, see LICENSE for details.
"""

__title__ = 'cmyui'
__author__ = 'cmyui'
__license__ = 'MIT'
__copyright__ = 'Copyright 2020 cmyui'
__version__ = '1.6.2'

from .logging import *
from .mysql import *
from .postgres import *
from .utils import *
from .version import *
from .web import *



# -*- coding: utf-8 -*-

"""\
Wrappers for multiple types of Discord related files.

Source: https://github.com/cmyui/cmyui_pkg

:copyright: (c) 2020 cmyui
:license: MIT, see LICENSE for details.
"""

from .webhook import *



# -*- coding: utf-8 -*-

import orjson
import aiohttp
from typing import Optional

__all__ = ('Footer', 'Image', 'Thumbnail',
           'Video', 'Provider', 'Author',
           'Field', 'Embed', 'Webhook')

class Footer:
    def __init__(self, text: str, **kwargs) -> None:
        self.text = text
        self.icon_url = kwargs.get('icon_url')
        self.proxy_icon_url = kwargs.get('proxy_icon_url')

class Image:
    def __init__(self, **kwargs) -> None:
        self.url = kwargs.get('url')
        self.proxy_url = kwargs.get('proxy_url')
        self.height = kwargs.get('height')
        self.width = kwargs.get('width')

class Thumbnail:
    def __init__(self, **kwargs) -> None:
        self.url = kwargs.get('url')
        self.proxy_url = kwargs.get('proxy_url')
        self.height = kwargs.get('height')
        self.width = kwargs.get('width')

class Video:
    def __init__(self, **kwargs) -> None:
        self.url = kwargs.get('url')
        self.height = kwargs.get('height')
        self.width = kwargs.get('width')

class Provider:
    def __init__(self, **kwargs) -> None:
        self.url = kwargs.get('url')
        self.name = kwargs.get('name')

class Author:
    def __init__(self, **kwargs) -> None:
        self.name = kwargs.get('name')
        self.url = kwargs.get('url')
        self.icon_url = kwargs.get('icon_url')
        self.proxy_icon_url = kwargs.get('proxy_icon_url')

class Field:
    def __init__(self, name: str, value: str,
                 inline: bool = False) -> None:
        self.name = name
        self.value = value
        self.inline = inline

class Embed:
    def __init__(self, **kwargs) -> None:
        self.title = kwargs.get('title')
        self.type = kwargs.get('type')
        self.description = kwargs.get('description')
        self.url = kwargs.get('url')
        self.timestamp = kwargs.get('timestamp') # datetime
        self.color = kwargs.get('color', 0x000000)

        self.footer: Optional[Footer] = kwargs.get('footer')
        self.image: Optional[Image] = kwargs.get('image')
        self.thumbnail: Optional[Thumbnail] = kwargs.get('thumbnail')
        self.video: Optional[Video] = kwargs.get('video')
        self.provider: Optional[Provider] = kwargs.get('provider')
        self.author: Optional[Author] = kwargs.get('author')

        self.fields: list[Field] = kwargs.get('fields', [])

    def set_footer(self, **kwargs) -> None:
        self.footer = Footer(**kwargs)

    def set_image(self, **kwargs) -> None:
        self.image = Image(**kwargs)

    def set_thumbnail(self, **kwargs) -> None:
        self.thumbnail = Thumbnail(**kwargs)

    def set_video(self, **kwargs) -> None:
        self.video = Video(**kwargs)

    def set_provider(self, **kwargs) -> None:
        self.provider = Provider(**kwargs)

    def set_author(self, **kwargs) -> None:
        self.author = Author(**kwargs)

    def add_field(self, name: str, value: str,
                  inline: bool = False) -> None:
        self.fields.append(Field(name, value, inline))

class Webhook:
    """A class to represent a single-use Discord webhook."""
    __slots__ = ('url', 'content', 'username',
                 'avatar_url', 'tts', 'file', 'embeds')
    def __init__(self, url: str, **kwargs) -> None:
        self.url = url
        self.content = kwargs.get('content')
        self.username = kwargs.get('username')
        self.avatar_url = kwargs.get('avatar_url')
        self.tts = kwargs.get('tts')
        self.file = kwargs.get('file')
        self.embeds = kwargs.get('embeds', [])

    def add_embed(self, embed: Embed) -> None:
        self.embeds.append(embed)

    @property
    def json(self):
        if not any([self.content, self.file, self.embeds]):
            raise Exception('Webhook must contain atleast one '
                            'of (content, file, embeds).')

        if self.content and len(self.content) > 2000:
            raise Exception('Webhook content must be under '
                            '2000 characters.')

        payload = {'embeds': []}

        for key in ('content', 'username',
                    'avatar_url', 'tts', 'file'):
            if (val := getattr(self, key)) is not None:
                payload[key] = val

        for embed in self.embeds:
            embed_payload = {}

            # simple params
            for key in ('title', 'type', 'description',
                        'url', 'timestamp', 'color'):
                if val := getattr(embed, key):
                    embed_payload[key] = val

            # class params, must turn into dict
            for key in ('footer', 'image', 'thumbnail',
                        'video', 'provider', 'author'):
                if val := getattr(embed, key):
                    embed_payload[key] = val.__dict__

            if embed.fields:
                embed_payload['fields'] = [f.__dict__ for f in embed.fields]

            payload['embeds'].append(embed_payload)

        return orjson.dumps(payload).decode()

    async def post(self, http: Optional[aiohttp.ClientSession] = None) -> None:
        """Post the webhook in JSON format."""
        _http = http or aiohttp.ClientSession(json_serialize=orjson.dumps)

        # TODO: if `self.file is not None`, then we should
        #       use multipart/form-data instead of json payload.
        headers = {'Content-Type': 'application/json'}
        async with _http.post(self.url, data=self.json,
                              headers=headers) as resp:
            if not resp or resp.status != 204:
                return # failed

        if not http:
            await _http.close()



# -*- coding: utf-8 -*-

from enum import IntFlag, unique

__all__ = ('Mods',)

@unique
class Mods(IntFlag):
    NOMOD       = 0
    NOFAIL      = 1 << 0
    EASY        = 1 << 1
    TOUCHSCREEN = 1 << 2 # old: 'NOVIDEO'
    HIDDEN      = 1 << 3
    HARDROCK    = 1 << 4
    SUDDENDEATH = 1 << 5
    DOUBLETIME  = 1 << 6
    RELAX       = 1 << 7
    HALFTIME    = 1 << 8
    NIGHTCORE   = 1 << 9
    FLASHLIGHT  = 1 << 10
    AUTOPLAY    = 1 << 11
    SPUNOUT     = 1 << 12
    AUTOPILOT   = 1 << 13
    PERFECT     = 1 << 14
    KEY4        = 1 << 15
    KEY5        = 1 << 16
    KEY6        = 1 << 17
    KEY7        = 1 << 18
    KEY8        = 1 << 19
    FADEIN      = 1 << 20
    RANDOM      = 1 << 21
    CINEMA      = 1 << 22
    TARGET      = 1 << 23
    KEY9        = 1 << 24
    KEYCOOP     = 1 << 25
    KEY1        = 1 << 26
    KEY3        = 1 << 27
    KEY2        = 1 << 28
    SCOREV2     = 1 << 29
    MIRROR      = 1 << 30

    # XXX: needs some modification to work..
    #KEY_MOD = KEY1 | KEY2 | KEY3 | KEY4 | KEY5 | KEY6 | KEY7 | KEY8 | KEY9 | KEYCOOP
    #FREE_MOD_ALLOWED = NOFAIL | EASY | HIDDEN | HARDROCK | \
    #                 SUDDENDEATH | FLASHLIGHT | FADEIN | \
    #                 RELAX | AUTOPILOT | SPUNOUT | KEY_MOD
    #SCORE_INCREASE_MODS = HIDDEN | HARDROCK | DOUBLETIME | FLASHLIGHT | FADEIN
    SPEED_CHANGING = DOUBLETIME | NIGHTCORE | HALFTIME

    def __repr__(self) -> str:
        if self.value == Mods.NOMOD:
            return ''

        mod_dict = {
            Mods.NOFAIL: 'NF',
            Mods.EASY: 'EZ',
            Mods.TOUCHSCREEN: 'TD',
            Mods.HIDDEN: 'HD',
            Mods.HARDROCK: 'HR',
            Mods.SUDDENDEATH: 'SD',
            Mods.DOUBLETIME: 'DT',
            Mods.RELAX: 'RX',
            Mods.HALFTIME: 'HT',
            Mods.NIGHTCORE: 'NC',
            Mods.FLASHLIGHT: 'FL',
            Mods.AUTOPLAY: 'AU',
            Mods.SPUNOUT: 'SO',
            Mods.AUTOPILOT: 'AP',
            Mods.PERFECT: 'PF',
            Mods.KEY4: 'K4',
            Mods.KEY5: 'K5',
            Mods.KEY6: 'K6',
            Mods.KEY7: 'K7',
            Mods.KEY8: 'K8',
            Mods.FADEIN: 'FI',
            Mods.RANDOM: 'RN',
            Mods.CINEMA: 'CN',
            Mods.TARGET: 'TP',
            Mods.KEY9: 'K9',
            Mods.KEYCOOP: 'CO',
            Mods.KEY1: 'K1',
            Mods.KEY3: 'K3',
            Mods.KEY2: 'K2',
            Mods.SCOREV2: 'V2',
            Mods.MIRROR: 'MR'
        }

        mod_str = []

        for m in (_m for _m in Mods if self.value & _m
                                    and _m != Mods.SPEED_CHANGING):
            mod_str.append(mod_dict[m])
        return ''.join(mod_str)

    @staticmethod
    def filter_invalid_combos(m: 'Mods') -> 'Mods':
        """Remove any invalid mod combinations from and return `m`."""
        if m & (Mods.DOUBLETIME | Mods.NIGHTCORE) and m & Mods.HALFTIME:
            m &= ~Mods.HALFTIME
        if m & Mods.EASY and m & Mods.HARDROCK:
            m &= ~Mods.HARDROCK
        if m & Mods.RELAX and m & Mods.AUTOPILOT:
            m &= ~Mods.AUTOPILOT
        if m & Mods.PERFECT and m & Mods.SUDDENDEATH:
            m &= ~Mods.SUDDENDEATH

        return m

    @classmethod
    def from_str(cls, s: str):
        # from fmt: `HDDTRX`
        # TODO: check for invalid mod combos
        mod_dict = {
            'EZ': cls.EASY,
            'NF': cls.NOFAIL,
            'HD': cls.HIDDEN,
            'PF': cls.PERFECT,
            'SD': cls.SUDDENDEATH,
            'HR': cls.HARDROCK,
            'NC': cls.NIGHTCORE,
            'DT': cls.DOUBLETIME,
            'HT': cls.HALFTIME,
            'FL': cls.FLASHLIGHT,
            'SO': cls.SPUNOUT,
            'RX': cls.RELAX,
            'AP': cls.AUTOPILOT
        }

        mods = cls.NOMOD

        for m in map(lambda i: s[i:i+2].upper(), range(0, len(s), 2)):
            if m not in mod_dict:
                continue

            mods |= mod_dict[m]

        return cls.filter_invalid_combos(mods)

    @classmethod
    def from_np(cls, s: str):
        # TODO: check for invalid mod combos
        # from fmt: `-DiffDown +DiffUp ~Special~`
        mod_dict = {
            '-Easy': cls.EASY,
            '-NoFail': cls.NOFAIL,
            '+Hidden': cls.HIDDEN,
            '+Perfect': cls.PERFECT,
            '+SuddenDeath': cls.SUDDENDEATH,
            '+HardRock': cls.HARDROCK,
            '+Nightcore': cls.NIGHTCORE,
            '+DoubleTime': cls.DOUBLETIME,
            '-HalfTime': cls.HALFTIME,
            '+Flashlight': cls.FLASHLIGHT,
            '-SpunOut': cls.SPUNOUT,
            '~Relax~': cls.RELAX,
            '~Autopilot~': cls.AUTOPILOT
        }

        mods = cls.NOMOD

        for mod in s.split(' '):
            # a bit unsafe.. perhaps defaultdict?
            mods |= mod_dict[mod]

        return cls.filter_invalid_combos(mods)



# -*- coding: utf-8 -*-

import aiomysql
from typing import Sequence, Any
from mysql.connector.pooling import MySQLConnectionPool
from typing import Optional, AsyncGenerator

__all__ = (
    # Informational
    'SQLParams',
    'SQLResult',

    # Functional
    'SQLPool',
    'AsyncSQLPool'
)

SQLParams = Sequence[Any]
SQLResult = Optional[dict[str, Any]]

class SQLPool:
    __slots__ = ('conn',)

    def __init__(self, **kwargs):
        self.conn = MySQLConnectionPool(autocommit=True, **kwargs)

    def execute(self, query: str, params: SQLParams = []) -> int:
        if not (cnx := self.conn.get_connection()):
            raise Exception('MySQL: Failed to retrieve a worker.')

        cur = cnx.cursor()
        cur.execute(query, params)
        cur.fetchmany()

        # Since we are executing a command, we
        # simply return the last row affected's id.
        res = cur.lastrowid

        cur.close()
        cnx.close()
        return res

    def fetch(self, query: str, params: SQLParams = [],
              _all: bool = False, _dict: bool = True) -> SQLResult:
        if not (cnx := self.conn.get_connection()):
            raise Exception('MySQL: Failed to retrieve a worker.')

        cur = cnx.cursor(dictionary=_dict, buffered=True)
        cur.execute(query, params)

        # We are fetching data.
        res = (cur.fetchall if _all else cur.fetchone)()

        cur.close()
        cnx.close()
        return res

    def fetchall(self, query: str, params: SQLParams = [],
                 _dict: bool = True ) -> tuple[SQLResult, ...]:
        return self.fetch(query, params, _all=True, _dict=_dict)

class AsyncSQLPool:
    __slots__ = ('pool',)

    def __init__(self):
        self.pool: Optional[aiomysql.Pool] = None

    async def connect(self, config):
        self.pool = await aiomysql.create_pool(**config, autocommit=True)

    async def close(self) -> None:
        self.pool.close()
        await self.pool.wait_closed()

    async def execute(self, query: str, params: SQLParams = []) -> int:
        """Acquire a connection & execute a query with params."""
        async with self.pool.acquire() as conn:
            async with conn.cursor(aiomysql.Cursor) as cur:
                await cur.execute(query, params)
                await conn.commit()

                return cur.lastrowid

    async def fetch(self, query: str, params: SQLParams = [],
                    _all: bool = False, _dict: bool = True
                   ) -> SQLResult:
        """Acquire a connection & execute query with params & fetch resultset(s)."""
        cursor_type = aiomysql.DictCursor if _dict \
                 else aiomysql.Cursor

        async with self.pool.acquire() as conn:
            async with conn.cursor(cursor_type) as cur:
                await cur.execute(query, params)
                return await (cur.fetchall if _all else cur.fetchone)()

    async def fetchall(self, query: str, params: SQLParams = [],
                       _dict: bool = True) -> tuple[SQLResult, ...]:
        return await self.fetch(query, params, _all=True, _dict=_dict)

    async def iterall(self, query: str, params: SQLParams = [],
                      _dict: bool = True) -> AsyncGenerator[SQLResult, None]:
        """Like fetchall, but as an async generator."""
        cursor_type = aiomysql.DictCursor if _dict \
                 else aiomysql.Cursor

        async with self.pool.acquire() as conn:
            async with conn.cursor(cursor_type) as cur:
                await cur.execute(query, params)

                async for row in cur:
                    yield row



# -*- coding: utf-8 -*-

__all__ = ('Version',)

class Version: # TODO: why fixed length? also could maybe inherit comp ops
    __slots__ = ('major', 'minor', 'micro')
    def __init__(self, major: int, minor: int, micro: int) -> None:
        self.major = major
        self.minor = minor
        self.micro = micro

    def __repr__(self) -> str:
        return '{0}.{1}.{2}'.format(*self.as_tuple)

    def __hash__(self) -> int:
        return self.as_tuple.__hash__()

    def __eq__(self, other: 'Version') -> bool:
        return self.as_tuple == other.as_tuple

    def __ne__(self, other: 'Version') -> bool:
        return self.as_tuple != other.as_tuple

    def __lt__(self, other: 'Version') -> bool:
        return self.as_tuple < other.as_tuple

    def __le__(self, other: 'Version') -> bool:
        return self.as_tuple <= other.as_tuple

    def __gt__(self, other: 'Version') -> bool:
        return self.as_tuple > other.as_tuple

    def __ge__(self, other: 'Version') -> bool:
        return self.as_tuple >= other.as_tuple

    @property
    def as_tuple(self) -> tuple[int]:
        return (self.major, self.minor, self.micro)

    @classmethod
    def from_str(cls, s: str):
        if len(split := s.split('.')) != 3:
            return

        return cls(*map(int, split))
